; aoaProcs.inc
;
; A set of macros for specifying type
; information about MASM variables.

            ifndef  aoaProcs_inc
aoaProcs_inc=       1

; locals macro syntax:
;
;   locals procName, comman-separated-list-of-declarations
;
; Each declaration takes the form:
;
;   id:type
;   id:type[expr]
;   align:type      (must be byte, word, dword, or qword)
;
; For each id declaraction, this macro generates two symbols
; of the form:
;
; procName$id$ofs =       offset
; procName$id     textequ <type ptr [rbp-offset]>
;
; where "offset" is the offset into the activation record
; (computed by this macro).
;
; The align declaration adjusts the offset so that it
; is appropriately aligned for the specified type's size
; (type should be byte, word, dword, or qword for align). 


locals      macro   procedure,items:vararg
            local   curOffset, curSize

curOffset   =       0       ;Offset into activation record

; For each argument on the command line:

            for     item,<items>
            local   colonPosn, id, arrayType
            local   lbrktPosn, rbrktPosn, symbol
            local   txteqt, typeqt, leneqt
            local   ofseqt, offs, tsizeqt, sizeqt
            local   nameqt

; Expecting "id:type", "id:type[expr]", or "align:type"
;
; Extract "id", "type", and optional "expr":
            
colonPosn   instr   1, <item>,<:>
id          substr  <item>,1, colonPosn-1 
arrayType   substr  <item>, colonPosn+1
lbrktPosn   instr   arrayType, <[>
rbrktPosn   instr   arrayType, <]>


            if      colonPosn eq 0
            echo    ***********************
%           echo    Expected a ":" after item
            echo    ***********************
            exitm
            endif

; If it was "align:type", adjust the current offset
; to match the alignment. Note that offsets are
; computed as positive values (even though they
; will eventually be negative).

            ifidn   id,<align>
            
curOffset   =       (curOffset + sizeof( arrayType ) - 1) and (-sizeof( arrayType ))

            else    ;We have a variable declaration
            
; See if there was an array declaration ("[expr]"). If so,
; extract the type and compute the number of elements.

            if      lbrktPosn ne 0
            
idType      substr  <item>, colonPosn+1, lbrktPosn-colonPosn+1
elements    substr  arrayType, lbrktPosn+1, rbrktPosn-lbrktPosn-1


            else

; If a scalar (non-array), extract the type name and
; set the number of elements to 1.

idType      textequ arrayType
elements    textequ <1>

            endif


; Compute the size of this variable and its offset into
; the activation record:

curSize     =       sizeof(idType) * (elements)
curOffset   =       curOffset + curSize


; Build the variable name as <procedure name> $ <id name>

symbol     catstr  <procedure>,<$>,id
offs        textequ %curOffset

; Generate equates for the offset and the variable:

ofseqt      catstr  <procedure>,<$>,id,<$ofs>,< = >, %(-curOffset)
txteqt      catstr  symbol,< textequ !<(>, idType,< ptr [rbp->, offs, <]) !>>
typeqt      catstr  symbol,<$typ textequ !<>, idType,<!>>
leneqt      catstr  symbol,<$elems =>, elements
tsizeqt     catstr  symbol,<$tsize =>, %sizeof( idType )
sizeqt      catstr  symbol,<$size =>, %curSize
nameqt      catstr  symbol,<$name textequ !<>, id, <!>>

            ofseqt
            txteqt
            typeqt
            leneqt
            tsizeqt
            sizeqt
            nameqt
            
            endif

            endm    ;for
            endm    ;locals
            
            
; parms macro syntax:
;
;   parms procName, offset, comman-separated-list-of-declarations
;
; Each declaration takes the form:
;
;   id:type
;   id:type[expr]
;
; For each id declaraction, this macro generates two symbols
; of the form:
;
; procName$id$ofs =       offset
; procName$id     textequ <type ptr [rbp+offset]>
;
; where "offset" is the offset into the activation record
; (computed by this macro).


parms       macro   procedure,baseReg,startOfs,items:vararg
            local   curOffset, curSize

curOffset   =       startOfs        ;Offset into activation record

; For each argument on the command line:

            for     item,<items>
            local   colonPosn, id, arrayType
            local   lbrktPosn, rbrktPosn, symbol
            local   txteqt, typeqt, leneqt
            local   ofseqt, offs, tsizeqt, sizeqt
            local   nameqt

; Expecting "id:type" or "id:type[expr]"
;
; Extract "id", "type", and optional "expr":
            
colonPosn   instr   1, <item>,<:>
id          substr  <item>,1, colonPosn-1 
arrayType   substr  <item>, colonPosn+1
lbrktPosn   instr   arrayType, <[>
rbrktPosn   instr   arrayType, <]>

            if      colonPosn eq 0
            echo    ***********************
%           echo    Expected a ":" after item
            echo    ***********************
            exitm
            endif
            
; See if there was an array declaration ("[expr]"). If so,
; extract the type and compute the number of elements.

            if      lbrktPosn ne 0
            
idType      substr  <item>, colonPosn+1, lbrktPosn-colonPosn+1
elements    substr  arrayType, lbrktPosn+1, rbrktPosn-lbrktPosn-1

            else

; If a scalar (non-array), extract the type name and
; set the number of elements to 1.

idType      textequ arrayType
elements    textequ <1>

            endif



; Build the parameter name as <procedure name> $ <id name>

symbol      catstr  <procedure>,<$>,id
offs        textequ %curOffset

; Generate equates for the offset and the variable:

curSize     =       sizeof(idType) * (elements)

ofseqt      catstr  <procedure>,<$>,id,<$ofs>,< = >, %curOffset
txteqt      catstr  symbol,< textequ !<(>, idType, < ptr [baseReg+>, offs, <]) !>>
typeqt      catstr  symbol,<$typ textequ !<>, idType,<!>>
leneqt      catstr  symbol,<$elems =>, elements
tsizeqt     catstr  symbol,<$tsize =>, %sizeof( arrayType )
sizeqt      catstr  symbol,<$size =>, %curSize
nameqt      catstr  symbol,<$name textequ !<>, id, <!>>

            ofseqt
            txteqt
            typeqt
            leneqt
            tsizeqt
            sizeqt
            nameqt
            
            
; Compute the size of this parameter and its offset into
; the activation record. Note that curSize is always
; some multiple of 8.

curSize     =       (curSize + 7) and (-8)
curOffset   =       curOffset + curSize

            endm    ;for
            endm    ;parms
         
         
;--------------------------------------------------------------
;
; Thunk macros
;
; setTh-
;  Initialize a thunk variable with the code that follows
;  this call and the current RBP value. Note that this macro
;  moves the following thunk code (up to the endTh invocation)
;  out of the current code stream.
;
;  Example:
;
;           setTh   someThunkVar
;
;      <code for the thunk>
;
;           endTh

         
setTh       macro   thunkVar
            local   thunkProc
            lea     rcx, thunkProc
            mov     (thunk_t ptr (thunkVar)).procedure, rcx
            mov     (thunk_t ptr (thunkVar)).environment, rbp
            
thunkSeg    segment execute
thunkProc:
            endm

; endTh-
;
;  End a thunk sequence begun by setTh

endTh       macro
thunkSeg    ends
            endm

; copyTh-
;
;  Copies the value of one thunk variable to another.
            
copyTh      macro   dest, src
            mov     rcx, (thunk_t ptr (src)).environment
            mov     (thunk_t ptr (dest)).environment, rcx
            mov     rcx, (thunk_t ptr (src)).procedure
            mov     (thunk_t ptr (dest)).procedure, rcx
            endm    
            
            
            endif   ;aoaProc_inc
