; classes.asm
;
; Library code to support MASM classes

            option     casemap:none
            include    aoaTypes.inc
            include    aoaMacros.inc
            include    aoaClasses.inc
            include    aoalib.inc
            
            
            .code

;------------------------------------------------------
;
; generic$ClassArray$Constructor:
;
; Generic constructor for a class.
;
; Input arguments:
;
; RSI-      Pointer to object (NULL if this code should
;           allocate storage for the object on the
;           heap).
;
; RDI-      Pointer to the class' VMT.
;
; ECX-      Size of class, in bytes.
;
; EDX-      Number of array elements to process
;           (0 or 1 for scalar objects).
;
; Assumption: VMT field is at offset 0 in the object.

generic$ClassArray$Constructor proc
                  push  rax
                  push  rdx
                  test  rsi, rsi ;NULL?
                  jnz   notNULL
                       
; If RSI was NULL on entry, allocate storage for
; a "protoClass" object on the heap. The actual
; amount of storage to allocate depends on the
; value in EDX. If EDX is zero or one, this
; code allocates storage for a single object.
; If EDX is greater than 1, it allocates storage
; for an array of objects.

                  test edx, edx
                  jnz  notZero
                  inc  edx

notZero:          mov  eax, ecx     ;Size of class
                  imul eax, edx     ;# of elements
                  call mem$alloc
                  mov  rsi, rax
                  test rsi, rsi
                  jz   failure
                       
; Initialize the VMT pointer for each element
; of the array:

notNULL:          push rsi
                  inc  rdx
                  jmp  endOfLoop
                  
forEachElement:   mov  [rsi], rdi
                  add  rsi, rcx     ;Next element
endOfLoop:        dec  edx
                  jnz  forEachElement
                  
                  pop  rsi
failure:          pop  rdx
                  pop  rax
                  ret
generic$ClassArray$Constructor endp
                  


;------------------------------------------------------
;
; generic$Class$Constructor:
;
; Initializes scalar class objects.
;
; Input arguments:
;
; RSI-      Pointer to object (NULL if this code should
;           allocate storage for the object on the
;           heap).
;
; RDI-      Pointer to the class' VMT.
;
; ECX-      Size of class, in bytes.



generic$Class$Constructor proc
                  push  rax
                  test  rsi, rsi ;NULL?
                  jnz   notNULL
                       
; If RSI was NULL on entry, allocate storage for
; a "protoClass" object on the heap. The actual
; amount of storage to allocate depends on the
; value in EDX. If EDX is zero or one, this
; code allocates storage for a single object.
; If EDX is greater than 1, it allocates storage
; for an array of objects.

                  mov  eax, ecx     ;Size of class
                  call mem$alloc
                  mov  rsi, rax
                  test rsi, rsi
                  jz   failure
                       
; Initialize the VMT pointer for the object:

notNULL:          mov  [rsi], rdi
                  
failure:          pop  rax
                  ret
generic$Class$Constructor endp
                  
            end