
            include    aoalib.inc
	include	   aoaMacros.inc   
            includelib aoalib.lib   


; aoaProc namespace for macros, constants, and other
; objects associated with the procedure declarations.

aoaProc	macro	ID
            namespace (&ID, <Pascal>, <aoaProc_Pascal>)
            namespace (&ID, <C>, <aoaProc_C>)
            namespace (&ID, <WinAPI>, <aoaProc_WinAPI>)
            namespace (&ID, <procStk>, <aoaProc_procStk>)
	echo	*****************************
	echo    Error:unexpected aoaProc
%	echo     member identifier=ID
	echo	*****************************
	exitm	<.err>
	endm

; Some useful constants:

aoaProc_Pascal	=	1
aoaProc_C		=	2
aoaProc_WinAPI	=	3


; Global constants that control the code generation
; for the procedure macro:
;
; @callConv- specifies procedure calling convention.
;	 Legal values: aoaProc(Pascal), aoaProc(C),
;	 and aoaProc(WinAPI).
;
; @display-	True if procedure should generate code to
;	maintain a display.
;
; @frame-	True if procedure should generate code to
;	construct and deconstruct the stack frame.

@callConv		=	aoaProc(Pascal)
@display		=	1
@frame		=	1	 	


;-------------------------------------------------------
;
; aoaProc(procStk) is the compile-time stack that keeps
; track of nested procedures. This also happens to be
; the current procedure's lexical level (minus 1).
	
aoaProc_procStk	=	0


	
;-------------------------------------------------------

procedure	macro	procName:vararg
	local	fnName, parenPosn, rParen
	local	trim0, trim1, parm, IDPosn
	local	genDisplay, genFrame, callConv
	local	parms, options, pnLen, pLen 
	local	optResult, optRslt2
	local	theID, theType, txtElements
	local	theElements
	local	commaPosn, brktPosn, lexSym
	local	offsetSym, typeSym, elemsSym
	local	saveOffset, saveType, saveElems
	local	saveLex, id, saveID 
	
%echo pn = procName
; Default values for frame/display generation and
; calling convention come from global symbols:

genDisplay	=	@display
genFrame	=	@frame
callConv	=	@callConv


; Make sure they've supply a procedure name argument:

	ifb	<&procName>
	echo	************************************
	echo	The "procedure" declaration requires
	echo	a procedure name and (possbly empty)
	echo	argument list.
	echo	************************************
	.err
	exitm
	endif
	
; Extract all the text in the argument field up to 
; an opening "(" symbol.

parenPosn	instr	1, <&procName>, <(>

; If no parentheses (and possibly optional arguments)
; then complain:

	if	parenPosn eq 0
	echo	************************************
	echo	Expected "(<optional parms>)"
	echo	************************************
	.err
	exitm
	else

; Extract everything up to the "(" as the procedure
; name. However, delete any trailing spaces from the
; procedure name to prevent problems when matching the
; name in the begin and endproc macros.
;
; Would like to be able to just write the code
; aoa(push)(aoaProc_procStk, aoa(trimWS)(procName))
; Sadly, that confuses MASM. So the following
; builds a textequ string that contains this
; very string and then expands the string all
; at once so MASM can deal with it.

trim0	substr	<&procName>, 1, parenPosn-1
trim1	textequ	<aoa_trimWS(trim0)>
parm	catstr	<aoa>, <(push)>, <(aoaProc_procStk,trim1)>

; Save the name on aoaProc_procStk so we can compare it with
; the names passed to the begin and endproc macros.
; At this point, parm is a text value with the
; following textual information:
;
; aoa(push) (aoaProc_procStk, trimmed_procedure_name)

	parm

	endif

		
; Okay, process the procedure's parameters (if any):
;
; Locate the end of the parameters by searching for
; the ")" character:

rParen	instr	1, <&procName>, <)>

; Extract the parameters; keep the ")" to use as a 
; sentinel for the parameter list.

parms	substr	<&procName>, parenPosn+1, rParen-parenPosn
pnLen	sizestr <&procName>


; Now get any options appearing at the end of the
; string:

	if	rParen lt pnLen
	
options	substr	<&procName>, rParen+1

	else
	
options	textequ	<>

	endif
	
; Because we might need the options values while
; processing the parameters, process the options
; first.
;
; Deal with the @display and @noDisplay options:

optResult	instr	options, <@display>
	if	optResult ne 0
genDisplay	=	1
optResult	instr	options, <@noDisplay>
	if	optResult ne 0
	echo	*****************************
	echo	@display and @noDisplay are
	echo	mutually exclusive
	echo	*****************************
	endif
	else
	
optResult	instr	options, <@noDisplay>
	if	optResult ne 0
genDisplay	=	0		
	endif
	endif


optResult	instr	options, <@frame>
	if	optResult ne 0
genFrame	=	1
optResult	instr	options, <@noFrame>
	if	optResult ne 0
	echo	*****************************
	echo	@frame and @noFrame are
	echo	mutually exclusive
	echo	*****************************
	endif
	else
	
optResult	instr	options, <@noFrame>
	if	optResult ne 0
genFrame	=	0		
	endif
	endif

optResult	instr	options, <@C>
	if	optResult ne 0
callConv	=	aoaProc(C)
optResult	instr	options, <@Pascal>
optRslt2	instr	options, <@WinAPI>
	if	(optResult+optRslt2) ne 0
	echo	*****************************
	echo	@C, @Pascal and @WinAPI are
	echo	mutually exclusive
	echo	*****************************
	endif
	
	else
	
	
optResult	instr	options, <@Pascal>
	if	optResult ne 0
callConv	=	aoaProc(Pascal)
optResult	instr	options, <@C>
optRslt2	instr	options, <@WinAPI>
	if	(optResult+optRslt2) ne 0
	echo	*****************************
	echo	@C, @Pascal and @WinAPI are
	echo	mutually exclusive
	echo	*****************************
	endif
	
	else
		
optResult	instr	options, <@WinAPI>
	if	optResult ne 0
callConv	=	aoaProc(WinAPI)
optResult	instr	options, <@C>
optRslt2	instr	options, <@Pascal>
	if	(optResult+optRslt2) ne 0
	echo	*****************************
	echo	@C, @Pascal and @WinAPI are
	echo	mutually exclusive
	echo	*****************************
	endif
	endif
	endif
	endif
	

; Okay, now time to start processing the 
; parameter list. Parameters take the following 
; form:
;
;  name:type{[expression]}
;
; where "[expression] is optional (denoted by the 
; braces) and specifies an array of the specified 
; type.
;
; Parameters are separated by commas.  The whole 
; parameter list ends with a ")" character.

parms	textequ	aoa_delWS(parms)
pLen	sizestr	parms
	while	pLen gt 0
	
; Extract the identifier at the beginning of the
; parameter list. Begin by locating the first
; ':' character starting from the beginning
; of the string:

IDPosn	instr	parms, <:>
	if	IDPosn le 1
	
	echo	******************************
	echo	Syntax error in parameter list
	echo	Expected "id:type", got
%	echo	<parms>
	echo	*****************************
	.err
	exitm
	
	endif

		
theID	substr	parms, 1, IDPosn-1
theID	textequ	aoa_trimWS(theID)

; Skip the colon and get the type:

pLen	sizestr	parms
	if	pLen lt 2
	echo	******************************
	echo	Syntax error in parameter list
	echo	expected ":type" got
%	echo	<parms>
	echo	*****************************
	.err
	exitm
	endif
	
parms	substr	parms, IDPosn+1	;Everything after ":"
parms	textequ	aoa_delWS(parms)

; The type ID ends with one of ",", "[", or ")".

commaPosn	instr	parms, <,>
parenPosn	instr	parms, <)>
brktPosn	instr	parms, <[>

	if	commaPosn eq 0
commaPosn	=	parenPosn
	endif
	if	(brktPosn ne 0)
	if	(commaPosn eq 0) or (commaPosn gt brktPosn)
commaPosn	=	brktPosn
	endif
	endif
	
	if	commaPosn eq 0
	echo	******************************
	echo	Syntax error in parameter list
	echo	expected ",", ")", or "[" to
	echo	end type name. Got:
%	echo	<parms>
	echo	*****************************
	.err
	exitm
	endif
	

	if	(brktPosn gt 0)
	if	(brktPosn lt commaPosn)
commaPosn	=	brktPosn
	endif
	endif

	if	(parenPosn gt 0)
	if	(parenPosn lt commaPosn)
commaPosn	=	parenPosn
	endif
	endif
	
; commaPosn now marks the position of the delimiter
; character. Grab everything up to that character
; and strip any trailing spaces:

theType	substr	parms, 1, commaPosn-1
theType	textequ	aoa_trimWS(theType)


	if	commaPosn lt parenPosn
parms	substr	parms, commaPosn+1
	else
parms	textequ	<>
	endif


; If the ID ended with "[" character, then process
; the expression for the array element count:

	if	commaPosn eq brktPosn
	
brktPosn	instr	parms, <]>
	if	brktPosn eq 0
	echo	******************************
	echo	Syntax error in parameter list
	echo	expected "[expr]". Missing
	echo	ending "]" character Got:
%	echo	<parms>
	echo	*****************************
	.err
	exitm
	endif
	
; Okay, extract the array size expression:

txtElements	substr	parms, 1, brktPosn-1
parms	substr	parms, brktPosn+1

; Convert the expression to a simple integer const:

theElements	=	txtElements
txtElements	textequ	%theElements

; Okay, we should have a "," or a ")" at this point

commaPosn	instr	1, parms, <,>
parenPosn	instr	1, parms, <)>
	if	(commaPosn eq 0) 
	exitm
	endif
	if	(parenPosn lt commaPosn)
	exitm
	endif
	
; Skip past the comma, if it is present:

parms	substr	parms, commaPosn+1

	else
theElements	=	0
txtElements	textequ	<0>
	endif
	
; Need to create a set of identifiers associated
; with this parameter symbol:
;
; <id>$offset: Holds the offset into the stack frame.
; <id>$type:   Holds the string containing symbol's type.
; <id>$elems:  Holds number of array elements for this
;	   symbol (0 if scalar).
; <id>$lexLvl: Hold's symbol's lexical level.
;
; <id>$saveOffset: Saves previous lexlevel's offset.
; <id>$saveType:   Saves previous lexlevel's type.
; <id>$saveElems:  Saves previous lexlevel's elems.
; <id>$saveLex:    Saves previous symbol's lex level.

tmp	textequ	theID
offsetSym	catstr	tmp, <$offset>
typeSym	textequ	<&theID&$type>
elemsSym	textequ	<&theID&$elems>
lexSym	textequ	<&theID&$lexLvl>
saveOffset	textequ	<&theID&$saveOffset>
saveType	textequ	<&theID&$saveType>
saveElems	textequ	<&theID&$saveElems>
saveLex	textequ	<&theID&$saveLex>

%echo os=offsetSym
ttt	textequ	%@defined( &offsetSym )
%echo os2=ttt

	if	@defined(&theID&)
	if	not (@defined( &offsetSym ))
	
	echo	******************************
%	echo	Symbol theID has already
	echo	been defined as a non-local
	echo	or non-parameter symbol.
	echo	******************************
	exitm
	.err
	endif

; Preserve the old values (typically at a different
; lexical level).
	
&saveOffset& textequ <&offsetSym&>
&saveType&	 textequ <&typeSym&>
&saveElems&	 textequ <&elemsSym&>
&saveLex&	 textequ <&lexSym&>	

	else	;Symbol is undefined at this point.

; No previous information to preserve, just use empty
; strings.
	
&saveOffset& textequ <>
&saveType&	 textequ <>
&saveElems&	 textequ <>
&saveLex&	 textequ <>	

	endif

%echo type=typeSym
&typeSym&	textequ	theType
&elemsSym&	textequ	txtElements
&lexSym	textequ	%aoaProc_procStk
	
	
		
; Repeat for the next parameter.

pLen	sizestr	parms
	endm
	
; Process the parameter list.

	endm



; begin macro-
;
; We're at the end of the parameter and local variable
; declarations. Time to start handling machine
; instructions.
	
begin	macro	beginName
	local	procName
	
; Verify that the name on the top of aoaProc_procStk matches
; the name following the begin:

procName	aoa_peek(aoaProc(procStk))
	ifdif	procName, <&beginName>
	echo	************************************
%	echo	Error: ID after "begin" (beginName)
%	echo	must match procedure name (procName)
	echo	************************************
	.err
	exitm
	endif
	
	endm

; endproc-
;
; Macro that marks the end of a procedure.
	
endproc	macro	endname
	local	procName
	
; Verify that the name on the top of aoaProc_procStk matches
; the name following the endproc:

procName	aoa_pop	(aoaProc(procStk))
	ifdif	procName, <&endname>
	echo	************************************
%	echo	Error: ID after "endproc" (endname)
%	echo	must match procedure name (procName)
	echo	************************************
	.err
	exitm
	endif
	
	endm
	
	
; makeLocal-
;
; A macro used to create local variable/parameter 
; macros during assembly. Given an argument cont-
; macros a single identifier "id", this macro 
; generatesa functional macro "id" that accepts the
; following arguments:
;
; id() -	returns the text "(x ptr [rbp+ofs])"
; 	where "x" is the local variable's
;	type and "ofs" is the offset to the
;	variable in the stack frame.
;
; id(offset)- Returns the offset to the local variable
; 	  or parameter in the stack frame.
;
; id(type)-	Returns the type of symbol "id".
;
; id(lex)-	Returns the lex level of the variable
;	(the index into the display whose entry
;	points at the activation record containing
;	local variable "id"; zero if the symbol is
;	local to the current procedure).

;makeLocal	macro	id, ofs, typ, lexLvl
;	local	macDcl
;	
;
;id	macro	arg
;%echo parm=<arg>
;	ifb	<&arg>
;	exitm	<(&typ& ptr [rbp+(ofs)])>
;	endif
;	
;	ifidn	<&arg>,<offset>
;	exitm	<&ofs>
;	endif
;	
;	ifidn	<&arg>,<type>
;	exitm	<&typ>
;	endif
;	
;	ifidn	<&arg>,<lex>
;	exitm	<&lexLvl>
;	endif
;	
;	echo	***********************************
;	echo	Unexpected argument in id macro:
;%	echo	arg
;	echo	***********************************
;	.err
;	exitm	<>
;	endm
;
;	
;	endm
	

	.code
		
echo procedure

b	macro	arg
	ifb	<&arg>
	exitm	<(dword ptr [rbp+(ofs)])>
	endif
	
	ifidn	<&arg>,<offset>
	exitm	<-5>
	endif
	
	ifidn	<&arg>,<type>
	exitm	<dword>
	endif
	
	ifidn	<&arg>,<lex>
	exitm	<0>
	endif
	
	echo	***********************************
%	echo	Unexpected argument in id macro:
%	echo	arg
	echo	***********************************
	.err
	exitm	<>
	endm
	
	
	
mov eax, b()
x	b(type) 55

	purge	b
	
	mov	eax, 0

b	macro	arg
	ifb	<&arg>
	exitm	<(dword ptr [rbp+(ofs)])>
	endif
	
	ifidn	<&arg>,<offset>
	exitm	<+5>
	endif
	
	ifidn	<&arg>,<type>
	exitm	<dword>
	endif
	
	ifidn	<&arg>,<lex>
	exitm	<0>
	endif
	
	echo	***********************************
%	echo	Unexpected argument in id macro:
%	echo	arg
	echo	***********************************
	.err
	exitm	<>
	endm

;mov eax, b()
;
;	purge	b
	
	
procedure fn  ( b : byte, c:byte[2+3], d:word ) @Pascal @noFrame @display

echo begin
begin fn
echo endproc
endproc fn

	

	end