; types.inc
; A set of macros for specifying type
; information about MASM variables.

	ifndef	types_inc
types_inc	=	1

; types$len-
;
; A private macro that returns the length of the
; argument.

types$len	 macro	v:req
	 exitm	%lengthof &v
	 endm


; type$macro-
;
; A helper macro for the type-specific macros that
; handles most of the work for those macros. See
; the individual macros that invoke type$macro
; for more information:
		 
type$macro	macro	v:req, tNam:req, tSiz:req, elms
	local	elements
	
; Define "id$type" (where "id" is the name passed
; as the "v" argument) and "id$tSize" (the type
; size, which is the size of one element if this
; is an array).

&v&$type    textequ <&tNam>
&v&$tSize   textequ <&tSiz>


; The elms (elements) field in the type macros
; is optional. If it it absent, assume the
; declaration is part of one of the following:
;
; A static data declaration in some segment, e.g.,
;
; byte_t(b) byte     16 dup (?)
;
; A parameter declaration in a proc statement, e.g.,
;
; someProc  proc     int32_t(parm):dword
;
; A local variable declaration in a proc, e.g.,
;
; someProc	proc
;           local    uns32_t(array)[16]:dword

	ifb	 <&elms>

; Compute the number of array elements using the
; usual MASM function. This calls types$len to
; compute the element length of argument v:

&v&$elems   textequ <types$lengthofVar(&v)>

; Compute the size of the entire array (elements
; times type size):
 
&v&$size    textequ <&v&$tSize*types$len(&v)>


; If the elms operand is non-blank, then the
; macro's invoker has explicitly specified the
; number of array elements. This is typically
; used for declarations like the following:
;
; word_t(myArray,10) textequ <[rbp-16-10*2]>
;
; (declares a 10-element word array when not
; using MASM HLL-like local declarations).

	else	;elms operand is not blank

;; If zero elements, force to 1:
		 
elements	 =	 elms + ((elms eq 0) and 1)

&v&$elems   textequ %elms 
&v&$size    textequ <&v&$tSize*elements>

	endif
		 
	exitm   <&v>
	endm
	

; Type macros.
;
; Use these macros to process identifiers in variable
; declarations appearing in your program. A macro
; invocation of the form "type_t(id)" where "type_t"
; represents one of these macros and "id" represents
; an identifier will return "id" as its text
; (effectively removing the macro invocation from the
; line) and also set up several global symbols the
; assembly language code can use to obtain information
; about the declaration. 
;
; The typical usage of these macros is the following:
;
; In a data segment:
;
; byte_t(b) byte    ?
; word_t(w) word	16 dup (?)
;
; In a procedure parameter declaration:
;
; someProc	proc	dword_t(parm1):dword, \
;                      uns64_t(parm2):qword, \
;		   real8_t(parm3):real8
;
; In the local variable declarations of a procedure:
;
; someProc	proc
;           local	qword_t(ptrVar):qword
;           local   byte_t(buf)[64]:byte
;
; You can also use these type macros with manually-
; defined memory objects (using text equates) as
; follows:
;
; int32_t(i,0) textequ <dword ptr [rbp+16]>
;
; The second macro argument specifies the number of
; array elements (if any, use 0 if this is a scalar
; object).

	
real4_t	macro   variable:req, elements
	exitm	type$macro(&variable, real4, 4, <&elements>)
	endm
		 
real8_t	macro   variable:req, elements
	exitm	type$macro(&variable, real8, 8, <&elements>)
	endm
		 
real10_t	macro   variable:req, elements
	exitm	type$macro(&variable, real10, 10, <&elements>)
	endm
	
		 
int64_t	macro   variable:req, elements
	exitm	type$macro(&variable, int64, 8, <&elements>)
	endm

int32_t	macro   variable:req, elements
	exitm	type$macro(&variable, int32, 4, <&elements>)
	endm
		 
int16_t	macro   variable:req, elements
	exitm	type$macro(&variable, int16, 2, <&elements>)
	endm

int8_t	macro   variable:req, elements
	exitm	type$macro(&variable, int8, 1, <&elements>)
	endm
	
		 
uns64_t	macro   variable:req, elements
	exitm	type$macro(&variable, uns64, 8, <&elements>)
	endm

uns32_t	macro   variable:req, elements
	exitm	type$macro(&variable, uns32, 4, <&elements>)
	endm
		 
uns16_t	macro   variable:req, elements
	exitm	type$macro(&variable, uns16, 2, <&elements>)
	endm

uns8_t	macro   variable:req, elements
	exitm	type$macro(&variable, uns8, 1, <&elements>)
	endm
	
		 
qword_t	macro   variable:req, elements
	exitm	type$macro(&variable, qword, 8, <&elements>)
	endm

dword_t	macro   variable:req, elements
	exitm	type$macro(&variable, dword, 4, <&elements>)
	endm
		 
word_t	macro   variable:req, elements
	exitm	type$macro(&variable, word, 2, <&elements>)
	endm

byte_t	macro   variable:req, elements
	exitm	type$macro(&variable, byte, 1, <&elements>)
	endm
		 
; Some useful macros for testing type information		 


; isType(id, type)-
;
; Returns true if "id" is declared as a type
; and is of type "type". Returns false if "id"
; is not a declared type or the type is not
; "type".

isType	macro	v:req, t:req

; If id$type is not defined, this is not a type-defined
; symbol; so, return false:

	ifndef	&v&$type
	exitm	<0>
	endif

; If the id$type is defined, see if
; it is equal the the parameter type "t":
	
	ifidn	<&v&$type>,<&t>
	exitm	<1>
	endif
	exitm	<0>
	endm
	
            


;; isArray(id, type)-
;;
;; Returns true if "id" is declared as a type
;; and has 1 or more array elements (scalars
;; have 0 elements).
;
;isArray	macro	v:req, t:req
;
;; If id$type is not defined, this is not a type-defined
;; symbol; so, return false:
;
;	ifndef	&v&$type
;	exitm	<0>
;	endif
;
;; If the id$type is defined, see if
;; it is equal the the parameter type "t":
;	
;	ifidn	<&v&$type>,<&t>
;	exitm	<1>
;	endif
;	exitm	<0>
;	endm
	
            
	endif	;types_inc